@@ -14,12 +14,9 @@ mod test_readme {
 }
 
 
-use std::{fmt, path::PathBuf, process, str::FromStr};
-
-use cfg_if::cfg_if;
-
-use crate::config::get_env_var;
+use std::{cmp::Ordering, fmt, path::PathBuf, process};
 
+use crate::{config::get_env_var, os::OperatingSystem};
 
 //======================================
 // Types
@@ -80,13 +77,13 @@ pub struct AppVersion {
     major: u32,
     minor: u32,
     revision: u32,
-
     minor_revision: Option<u32>,
+
     build_code: u32,
 }
 
 /// Wolfram Language version number.
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
 #[non_exhaustive]
 pub struct WolframVersion {
     major: u32,
@@ -111,13 +108,25 @@ impl std::error::Error for Error {}
 
 /// Discover all installed Wolfram applications.
 ///
+/// The [`WolframApp`] elements in the returned vector will be sorted by Wolfram
+/// Language version and application feature set. The newest and most general app
+/// will be at the start of the list.
+///
 /// # Caveats
 ///
 /// This function will use operating-system specific logic to discover installations of
 /// Wolfram applications. If a Wolfram application is installed to a non-standard
 /// location, it may not be discoverable by this function.
 pub fn discover() -> Vec<WolframApp> {
-    os::discover_all()
+    let mut apps = os::discover_all();
+
+    // Sort `apps` so that the "best" app is the last element in the vector.
+    apps.sort_by(WolframApp::best_order);
+
+    // Reverse `apps`, so that the best come first.
+    apps.reverse();
+
+    apps
 }
 
 /// Discover all installed Wolfram applications that match the specified filtering
@@ -129,7 +138,7 @@ pub fn discover() -> Vec<WolframApp> {
 /// Wolfram applications. If a Wolfram application is installed to a non-standard
 /// location, it may not be discoverable by this function.
 pub fn discover_with_filter(filter: &Filter) -> Vec<WolframApp> {
-    let mut apps = os::discover_all();
+    let mut apps = discover();
 
     apps.retain(|app| filter.check_app(&app).is_ok());
 
@@ -148,17 +157,12 @@ pub fn discover_with_filter(filter: &Filter) -> Vec<WolframApp> {
 //         `as_str(&self) -> &'static str`
 //       method.
 pub fn target_system_id() -> &'static str {
-    cfg_if![
-        if #[cfg(all(target_os = "macos", target_arch = "x86_64"))] {
-            const SYSTEM_ID: &str = "MacOSX-x86-64";
-        } else {
-            // FIXME: Update this to include common Linux/Windows (and ARM macOS)
-            //        platforms.
-            compile_error!("target_system_id() has not been implemented for the current system")
-        }
-    ];
-
-    SYSTEM_ID
+    match system_id_from_target(env!("TARGET")) {
+        Ok(system_id) => system_id,
+        Err(err) => panic!(
+            "target_system_id() has not been implemented for the current target: {err}"
+        ),
+    }
 }
 
 /// Returns the System ID value that corresponds to the specified Rust
@@ -166,7 +170,17 @@ pub fn target_system_id() -> &'static str {
 /// any.
 pub fn system_id_from_target(rust_target: &str) -> Result<&'static str, Error> {
     let id = match rust_target {
+        // 64-bit x86
         "x86_64-apple-darwin" => "MacOSX-x86-64",
+        "x86_64-unknown-linux-gnu" => "Linux-x86-64",
+        "x86_64-pc-windows-msvc" | "x86_64-pc-windows-gnu" => "Windows-x86-64",
+        // 64-bit ARM
+        "aarch64-apple-darwin" => "MacOSX-ARM64",
+        "aarch64-apple-ios" | "aarch64-apple-ios-sim" => "iOS-ARM64", // iOS
+        "aarch64-linux-android" => "Android",
+
+        // 32-bit ARM (e.g. Raspberry Pi)
+        "armv7-unknown-linux-gnueabihf" => "Linux-ARM",
         _ => {
             return Err(Error(format!(
                 "no System ID value associated with Rust target triple: {}",
@@ -198,6 +212,42 @@ impl WolframAppType {
             WolframAlphaNotebookEdition,
         ]
     }
+
+    /// The 'usefulness' value of a Wolfram application type, all else being equal.
+    ///
+    /// This is a rough, arbitrary indicator of how general and flexible the Wolfram
+    /// Language capabilites offered by a particular application type are.
+    ///
+    /// This relative ordering is not necessarily best for all use cases. For example,
+    /// it will rank a Wolfram Engine installation above Wolfram Player, but e.g. an
+    /// application that needs a notebook front end may actually prefer Player over
+    /// Wolfram Engine.
+    //
+    // TODO: Break this up into separately orderable properties, e.g. `has_front_end()`,
+    //       `is_restricted()`.
+    #[rustfmt::skip]
+    fn ordering_value(&self) -> u32 {
+        use WolframAppType::*;
+
+        match self {
+            // Unrestricted | with a front end
+            Desktop => 100,
+            Mathematica => 99,
+            FinancePlatform => 98,
+            ProgrammingLab => 97,
+
+            // Unrestricted | without a front end
+            Engine => 96,
+
+            // Restricted | with a front end
+            PlayerPro => 95,
+            Player => 94,
+            WolframAlphaNotebookEdition => 93,
+
+            // Restricted | without a front end
+            // TODO?
+        }
+    }
 }
 
 impl WolframVersion {
@@ -441,14 +491,38 @@ impl WolframApp {
     /// Location of the application's main executable.
     ///
     /// * **macOS:** `CFBundleCopyExecutableURL()` location.
-    /// * **Windows:** *TODO*
+    /// * **Windows:** `RegGetValue(_, _, "ExecutablePath", ...)` location.
     /// * **Linux:** *TODO*
     pub fn app_executable(&self) -> Option<PathBuf> {
         self.app_executable.clone()
     }
 
-    /// Returns the version of the Wolfram Language bundled with this application.
+    /// Returns the version of the [Wolfram Language][WL] bundled with this application.
+    ///
+    /// [WL]: https://wolfram.com/language
     pub fn wolfram_version(&self) -> Result<WolframVersion, Error> {
+        if self.app_version.major == 0 {
+            return Err(Error(format!(
+                "wolfram app has invalid application version: {:?}  (at: {})",
+                self.app_version,
+                self.app_directory.display()
+            )));
+        }
+
+        // TODO: Are there any Wolfram products where the application version number is
+        //       not the same as the Wolfram Language version it contains?
+        //
+        //       What about any Wolfram apps that do not contain a Wolfram Languae instance?
+        Ok(WolframVersion {
+            major: self.app_version.major,
+            minor: self.app_version.minor,
+            patch: self.app_version.revision,
+        })
+
+        /* TODO:
+            Look into fixing or working around the `wolframscript` hang on Windows, and generally
+            improving this approach. E.g. use WSTP instead of parsing the stdout of wolframscript.
+
         // MAJOR.MINOR
         let major_minor = self
             .wolframscript_output("$VersionNumber")?
@@ -482,6 +556,7 @@ impl WolframApp {
             minor,
             patch,
         })
+        */
     }
 
     /// The [`$InstallationDirectory`][ref/$InstallationDirectory] of this Wolfram System
@@ -493,14 +568,16 @@ impl WolframApp {
             return player.installation_directory();
         }
 
-        if cfg!(target_os = "macos") {
-            self.app_directory.join("Contents")
-        } else {
-            // FIXME: Fill this in for Windows and Linux
-            panic!(
-                "{}",
-                platform_unsupported_error("WolframApp::from_app_directory()",)
-            )
+        match OperatingSystem::target_os() {
+            OperatingSystem::MacOS => self.app_directory.join("Contents"),
+            OperatingSystem::Windows => self.app_directory.clone(),
+            // FIXME: Fill this in for Linux
+            OperatingSystem::Linux | OperatingSystem::Other => {
+                panic!(
+                    "{}",
+                    platform_unsupported_error("WolframApp::from_app_directory()",)
+                )
+            },
         }
     }
 
@@ -512,14 +589,20 @@ impl WolframApp {
     /// [`WolframKernel`](https://reference.wolfram.com/language/ref/program/WolframKernel.html)
     /// executable.
     pub fn kernel_executable_path(&self) -> Result<PathBuf, Error> {
-        let path = if cfg!(target_os = "macos") {
-            // TODO: In older versions of the product, MacOSX was used instead of MacOS.
-            //       Look for either, depending on the version number.
-            self.installation_directory()
-                .join("MacOS")
-                .join("WolframKernel")
-        } else {
-            return Err(platform_unsupported_error("kernel_executable_path()"));
+        let path = match OperatingSystem::target_os() {
+            OperatingSystem::MacOS => {
+                // TODO: In older versions of the product, MacOSX was used instead of MacOS.
+                //       Look for either, depending on the version number.
+                self.installation_directory()
+                    .join("MacOS")
+                    .join("WolframKernel")
+            },
+            OperatingSystem::Windows => {
+                self.installation_directory().join("WolframKernel.exe")
+            },
+            OperatingSystem::Linux | OperatingSystem::Other => {
+                return Err(platform_unsupported_error("kernel_executable_path()"));
+            },
         };
 
         if !path.is_file() {
@@ -540,12 +623,14 @@ impl WolframApp {
             return player.wolframscript_executable_path();
         }
 
-        let path = if cfg!(target_os = "macos") {
-            PathBuf::from("MacOS").join("wolframscript")
-        } else {
-            return Err(platform_unsupported_error(
-                "wolframscript_executable_path()",
-            ));
+        let path = match OperatingSystem::target_os() {
+            OperatingSystem::MacOS => PathBuf::from("MacOS").join("wolframscript"),
+            OperatingSystem::Windows => PathBuf::from("wolframscript.exe"),
+            OperatingSystem::Linux | OperatingSystem::Other => {
+                return Err(platform_unsupported_error(
+                    "wolframscript_executable_path()",
+                ));
+            },
         };
 
         let path = self.installation_directory().join(&path);
@@ -580,10 +665,16 @@ impl WolframApp {
     /// [WSTP](https://reference.wolfram.com/language/guide/WSTPAPI.html)
     /// static library.
     pub fn wstp_static_library_path(&self) -> Result<PathBuf, Error> {
-        let static_archive_name = if cfg!(target_os = "macos") {
-            "libWSTPi4.a"
-        } else {
-            return Err(platform_unsupported_error("wstp_static_library_path()"));
+        let static_archive_name = match OperatingSystem::target_os() {
+            // Note: In theory, this can also vary based on the WSTP library 'interface' version
+            //       (currently v4). But that has not changed in a long time. If the interface
+            //       version does change, this logic should be updated to also check the WL
+            //       version.
+            OperatingSystem::MacOS => "libWSTPi4.a",
+            OperatingSystem::Windows => "wstp64i4s.lib",
+            OperatingSystem::Linux | OperatingSystem::Other => {
+                return Err(platform_unsupported_error("wstp_static_library_path()"));
+            },
         };
 
         let lib = self
@@ -622,12 +713,11 @@ impl WolframApp {
             return Ok(PathBuf::from(path));
         }
 
-        let path = if cfg!(target_os = "macos") {
-            self.installation_directory()
-                .join("SystemFiles/IncludeFiles/C/")
-        } else {
-            return Err(platform_unsupported_error("library_link_c_includes_path()"));
-        };
+        let path = self
+            .installation_directory()
+            .join("SystemFiles")
+            .join("IncludeFiles")
+            .join("C");
 
         if !path.is_dir() {
             return Err(Error(format!(
@@ -640,6 +730,62 @@ impl WolframApp {
         Ok(path)
     }
 
+    //----------------------------------
+    // Sorting `WolframApp`s
+    //----------------------------------
+
+    /// Order two `WolframApp`s by which is "best".
+    ///
+    /// This comparison will sort apps using the following factors in the given order:
+    ///
+    /// * Wolfram Language version number.
+    /// * Application feature set (has a front end, is unrestricted)
+    ///
+    /// For example, [Mathematica][WolframAppType::Mathematica] is a more complete
+    /// installation of the Wolfram System than [Wolfram Engine][WolframAppType::Engine],
+    /// because it provides a notebook front end.
+    ///
+    /// See also [WolframAppType::ordering_value()].
+    fn best_order(a: &WolframApp, b: &WolframApp) -> Ordering {
+        //
+        // First, sort by Wolfram Language version.
+        //
+
+        let version_order = match (a.wolfram_version().ok(), b.wolfram_version().ok()) {
+            (Some(a), Some(b)) => a.cmp(&b),
+            (Some(_), None) => Ordering::Greater,
+            (None, Some(_)) => Ordering::Less,
+            (None, None) => Ordering::Equal,
+        };
+
+        if version_order != Ordering::Equal {
+            return version_order;
+        }
+
+        //
+        // Then, sort by application type.
+        //
+
+        // Sort based roughly on the 'usefulness' of a particular application type.
+        // E.g. Wolfram Desktop > Mathematica > Wolfram Engine > etc.
+        let app_type_order = {
+            let a = a.app_type().ordering_value();
+            let b = b.app_type().ordering_value();
+            a.cmp(&b)
+        };
+
+        if app_type_order != Ordering::Equal {
+            return app_type_order;
+        }
+
+        debug_assert_eq!(a.wolfram_version().ok(), b.wolfram_version().ok());
+        debug_assert_eq!(a.app_type().ordering_value(), b.app_type().ordering_value());
+
+        // TODO: Are there any other metrics by which we could sort this apps?
+        //       Installation location? Released build vs Prototype/nightly?
+        Ordering::Equal
+    }
+
     //----------------------------------
     // Utilities
     //----------------------------------
@@ -653,14 +799,14 @@ impl WolframApp {
             return Ok(PathBuf::from(path));
         }
 
-        let path = if cfg!(target_os = "macos") {
-            self.installation_directory()
-                .join("SystemFiles/Links/WSTP/DeveloperKit/")
-                .join(target_system_id())
-                .join("CompilerAdditions")
-        } else {
-            return Err(platform_unsupported_error("wstp_compiler_additions_path()"));
-        };
+        let path = self
+            .installation_directory()
+            .join("SystemFiles")
+            .join("Links")
+            .join("WSTP")
+            .join("DeveloperKit")
+            .join(target_system_id())
+            .join("CompilerAdditions");
 
         if !path.is_dir() {
             return Err(Error(format!(
@@ -672,6 +818,7 @@ impl WolframApp {
         Ok(path)
     }
 
+    #[allow(dead_code)]
     fn wolframscript_output(&self, input: &str) -> Result<String, Error> {
         let mut args = vec!["-code".to_owned(), input.to_owned()];
 
@@ -700,6 +847,7 @@ pub(crate) fn print_platform_unimplemented_warning(op: &str) {
     )
 }
 
+#[cfg_attr(target_os = "windows", allow(dead_code))]
 fn warning(message: &str) {
     eprintln!("warning: {}", message)
 }
@@ -782,21 +930,25 @@ impl WolframApp {
             return Ok(self);
         }
 
-        let embedded_player_path = if cfg!(target_os = "macos") {
-            self.app_directory
+        let embedded_player_path = match OperatingSystem::target_os() {
+            OperatingSystem::MacOS => self
+                .app_directory
                 .join("Contents")
                 .join("Resources")
-                .join("Wolfram Player.app")
-        } else {
-            // TODO: Does Wolfram Engine on Linux/Windows contain an embedded Wolfram Player,
-            //       or is that only done on macOS?
-            print_platform_unimplemented_warning(
-                "determine Wolfram Engine path to embedded Wolfram Player",
-            );
+                .join("Wolfram Player.app"),
+            OperatingSystem::Windows
+            | OperatingSystem::Linux
+            | OperatingSystem::Other => {
+                // TODO: Does Wolfram Engine on Linux/Windows contain an embedded Wolfram Player,
+                //       or is that only done on macOS?
+                print_platform_unimplemented_warning(
+                    "determine Wolfram Engine path to embedded Wolfram Player",
+                );
 
-            // On the hope that returning `app` is more helpful than returning an error here,
-            // do that.
-            return Ok(self);
+                // On the hope that returning `app` is more helpful than returning an error here,
+                // do that.
+                return Ok(self);
+            },
         };
 
         // TODO: If this `?` propagates an error
